(load "env.scm")
(load "parser.scm")

(define-struct expression ())
(define-struct (lit-exp   expression) (value))
(define-struct (var-exp   expression) (name))
(define-struct (if-exp    expression) (test then else))
(define-struct (begin-exp expression) (exps))
(define-struct (let-exp   expression) (params args body))
(define-struct (fun-exp   expression) (params body))
(define-struct (app-exp   expression) (fun args))
(define-struct (prim-exp  expression) (prim args))

;; reflective access to current jp
(define-struct (this-jp-exp  expression) ())

;; aspect deployment
(define-struct (depl-exp  expression) (ds asp body))

;; closure values
(define-struct value ())
(define-struct (closure value) (params body env aspects))

;; aspects
(define-struct dasp (c d f pc adv))
(define-struct jp   (closure args parent))

(define root-jp (make-jp (make-closure '() '() (make-empty-env) '())
                         '() #f))

(define (jp-depth jp)
  (if (not jp)
      0
      (if (not (jp-parent jp))
          1
          (+ 1 (jp-depth (jp-parent jp))))))

(define (stack-depth-less n jp)
  (lambda (jp2)
    (<= n (- (jp-depth jp2) (jp-depth jp)))))

;; evaluate expression exp in lexical environment E and aspect environment A, with current join point jp
(define (eval exp E A jp)
  ;(printf "eval: ~a~n" exp)
  (cond ((lit-exp? exp) (lit-exp-value exp))
        
        ((prim-exp? exp)
         (let ((args (eval-args (prim-exp-args exp) E A jp)))
           (apply (prim-exp-prim exp) args)))
        
        ((if-exp? exp)
         (if (eval (if-exp-test exp) E A jp)
             (eval (if-exp-then exp) E A jp)
             (eval (if-exp-else exp) E A jp)))
        
        ((begin-exp? exp)
         (eval-body (begin-exp-exps exp) E A jp))
        
        ((let-exp? exp)
         (let ((args   (eval-args (let-exp-args exp) E A jp)))
           (eval-body (let-exp-body exp) (extend-env (let-exp-params exp) args E) A jp)))

        ((var-exp? exp)   (env-get (var-exp-name exp) E))
        
        ((depl-exp? exp)  (let ((dasp (build-dasp exp E A jp)))
                            (eval-body (depl-exp-body exp)  E (cons dasp A) jp)))
        
        ((fun-exp? exp)   (make-closure (fun-exp-params exp) (fun-exp-body exp) E (collect-match-d jp A)))
        
        ((app-exp? exp)   (let* ((cl   (eval (app-exp-fun exp) E A jp))
                                 (args (eval-args (app-exp-args exp) E A jp))
                                 (njp  (make-jp cl args jp))
                                 (env  (extend-env (closure-params cl) args (closure-env cl)))
                                 (asps (append (collect-match-c njp A) (closure-aspects cl))))
                            (weave-some A njp) 
                            (eval-body (closure-body cl) env asps njp
                                  )))
        ((this-jp-exp? exp) jp)
        ))

(define (eval-args exps E A jp)
    (map (lambda (x) (eval x E A jp)) exps))

(define (eval-body body E A jp)
  (let loop ((exps body) (val #f))
    (if (null? exps)
        val
        (loop (cdr exps) (eval (car exps) E A jp)))))

(define (build-dasp exp E A jp)
  (let ((ds  (eval (depl-exp-ds exp) E A jp))
        (asp (eval (depl-exp-asp exp) E A jp)))
    (make-dasp (car ds)
               (cadr ds)
               (caddr ds)
               (car asp)
               (cadr asp))))


;; aspect environments
(define (collect-match-c jp asps) (collect-if (lambda (a) (app/prim (dasp-c a) jp)) asps))
(define (collect-match-d jp asps) (collect-if (lambda (a) (app/prim (dasp-d a) jp)) asps))
(define (collect-match-f jp asps) (collect-if (lambda (a) (app/prim (dasp-f a) jp)) asps))

;; weaving
(define (weave-some asps jp)
  (weave-all (collect-match-f jp asps) jp))
   
(define (weave-all asps jp)
  (map (lambda (x) (weave x jp)) asps))

(define (weave asp jp)
  (if (app/prim (dasp-pc asp) jp)
      (app/prim (dasp-adv asp) jp)))

;; applies the function f, passing jp as argument
;; evaluation is done in an empty aspect environment
(define (app/prim f jp)
  (eval-body (closure-body f)
             (extend-env (closure-params f) (list jp)
                         (closure-env f)) '() jp))
