(load "ds-fp.scm")

;;; basic tests
(eval (parse '10) '() '() root-jp)
(eval (parse 'x) (new-env (list 'x) (list 15)) '() root-jp)
(eval (parse '(+ 1 x)) (new-env (list 'x) (list 15)) '() root-jp)
(eval (parse '(fun (x y) (+ x y))) (make-empty-env) '() root-jp)
(eval (parse '(let ((y 10)) (+ y x))) (new-env (list 'x) (list 15)) '() root-jp)
(eval (parse '(let ((f (fun (x y) (+ x y))))
          (write (f 3 4))
          (f 1 1))) (make-empty-env) '() root-jp)


;; Expressing AspectScheme semantics with deployment strategies
(printf "~n---dynamically-scoped: aspect applies~n")
(eval (parse 
       '(let ((t (fun (jp) #t))
              (f (fun (jp) #f))
              (call (fun (f) (fun (jp) (eq-jp? jp f))))
              (open-file (fun (file) (printf "->opening ~a~n" file))))
          (let ((apply-to-brussels (fun (f) (f "brussels")))) 
            (deploy (list t f t) 
                    (list (call open-file) (fun (jp) (printf "trace~n")))
                    (apply-to-brussels open-file))))
       ) (make-empty-env) '() root-jp)

(printf "~n---statically-scoped: aspect does not apply~n")
(eval (parse 
       '(let ((t (fun (jp) #t))
              (f (fun (jp) #f))
              (call (fun (f) (fun (jp) (eq-jp? jp f))))
              (open-file (fun (file) (printf "->opening ~a~n" file))))
          (let ((apply-to-brussels (fun (f) (f "brussels"))))
            (deploy (list f t t) 
                    (list (call open-file) (fun (jp) (printf "trace~n")))
                    (apply-to-brussels open-file))))
       ) (make-empty-env) '() root-jp)

(printf "~n---statically-scoped: aspect applies~n")
(eval (parse
       '(let ((t (fun (jp) #t))
              (f (fun (jp) #f))
              (call (fun (f) (fun (jp) (eq-jp? jp f))))
              (open-file (fun (file) (printf "->opening ~a~n" file))))
          (let ((traced-open  (deploy (list f t t) 
                                      (list (call open-file) (fun (jp) (printf "trace~n")))
                                      (fun (f) (open-file f)))))
              (traced-open "brussels")))
          ) (make-empty-env) '() root-jp)

(printf "~n---dynamically-scoped: aspect does not apply~n")
(eval (parse
       '(let ((t (fun (jp) #t))
              (f (fun (jp) #f))
              (call  (fun (f) (fun (jp) (eq-jp? jp f))))
              (open-file (fun (file) (printf "->opening ~a~n" file))))
          (let ((traced-open (deploy (list t f t) 
                                      (list (call open-file) (fun (jp) (printf "trace~n")))
                                      (fun (f) (open-file f)))))
            (traced-open "brussels")))
       ) (make-empty-env) '() root-jp)

;;; parameterized propagation functions
(printf "~n---propagate on stack only if applying apply-to-paris: once~n")
(eval (parse 
       '(let ((t (fun (jp) #t))
              (f (fun (jp) #f))
              (call  (fun (f) (fun (jp) (eq-jp? jp f))))
              (open-file (fun (file) (printf "->opening ~a~n" file))))
          (let ((apply-to-brussels (fun (f) (f "brussels")))
                (apply-to-paris    (fun (f) (f "paris")))) 
            (deploy (list (fun (jp) (eq-jp? jp apply-to-paris)) t t) 
                    (list (fun (jp) (eq-jp? jp open-file)) (fun (jp) (printf "trace~n")))
                    (apply-to-brussels open-file)
                    (apply-to-paris open-file)
                    )))
       ) (make-empty-env) '() root-jp)

(printf "~n---propagate on stack depending on relative stack depth~n")
(eval (parse
       '(let ((t (fun (jp) #t))
              (f (fun (jp) #f))
              (call (fun (f) (fun (jp) (eq-jp? jp f))))
              (stack-depth-<= (fun (n jp) 
                                     (fun (jp2) 
                                          (<= (- (jp-depth jp2) (jp-depth jp)) n))))
                                               
              (open-file (fun (file) (printf "->opening ~a~n" file))))
          (let ((open-log (fun () (open-file "log"))))
            (let ((open-both (fun (file) 
                                  (open-file file)
                                  (open-log))))
              (deploy (list (stack-depth-<= 1 (this-jp)) f t) 
                      (list (call open-file) (fun (jp) (printf "trace~n")))
                      (open-both "brussels")))))
            ) (make-empty-env) '() root-jp)


;;; join point filters
(printf "~n---join point filter~n")
(eval (parse 
       '(let ((t (fun (jp) #t))
              (f (fun (jp) #f))
              (call  (fun (f) (fun (jp) (eq-jp? jp f))))
              (open-file (fun (file) (printf "->opening ~a~n" file)))
              (aspect (list (fun (jp) #t) (fun (jp) (printf "trace~n"))))) ;; trace all applications
          
          (let ((apply-to-brussels (fun (f) (f "brussels")))
                (apply-to-paris    (fun (f) (f "paris")))) 
            
            (printf "here match twice~n")
            (deploy (list t t t) 
                    aspect
                    (apply-to-paris open-file))
            
            (printf "here match once~n")
            (deploy (list t t (fun (jp) (eq-jp? jp apply-to-paris))) 
                    aspect 
                    (apply-to-paris open-file))))
            
       ) (make-empty-env) '() root-jp)